-- | `App.DomainListInterface` is a simple component with the list of own domains
-- | and a form to add a new domain.
-- |
-- | This interface allows to:
-- |   - display the list of own domains
-- |   - show and select accepted domains (TLDs)
-- |   - create new domains
-- |   - delete a domain
-- |   - ask for confirmation
-- |   - switch to the interface to show and modify the content of a Zone
-- |   - TODO: validate the domain before sending a message to `dnsmanagerd`
module App.Page.DomainList where

import Prelude (Unit, bind, discard, map, otherwise, pure, ($), (/=), (<<<), (<>), (>))

import Data.Array as A
import Data.ArrayBuffer.Types (ArrayBuffer)
import Data.Either (Either(..))
import Data.Maybe (Maybe(..), maybe)
import Data.String.Utils (endsWith)
import Effect.Aff.Class (class MonadAff)
import Halogen as H
import Halogen.HTML as HH
import Halogen.HTML.Events as HE
import Halogen.HTML.Events as HHE
import Halogen.HTML.Properties as HP
import Web.Event.Event as Event
import Web.Event.Event (Event)
import Bulma as Bulma

import App.DisplayErrors (error_to_paragraph_label)

import App.Validation.Label as Validation

import CSSClasses as C
import App.Type.LogMessage
import App.Message.DNSManagerDaemon as DNSManager

-- | `App.DomainListInterface` can send messages through websocket interface
-- | connected to dnsmanagerd. See `App.WS`.
-- |
-- | Also, this component can log messages and ask its parent (`App.Container`) to
-- | reconnect the websocket to `dnsmanagerd`.
-- |
-- | Finally, the component can ask its state to its parent.
-- | The reason is quite simple.
-- | The component can be deleted, meaning that it loses its state.
-- | Instead of asking `dnsmanagerd` the list of available domains and the list of owned domains
-- | each time the component is instanciated, the parent stores the component's state when the
-- | component is removed. This way, the data is conserved.

data Output
  = MessageToSend ArrayBuffer
  | Log LogMessage
  | ChangePageZoneInterface String
  | AskState
  | StoreState State

-- | `App.DomainListInterface` can receive messages from `dnsmanagerd`.
-- |
-- | The component is also informed when the connection is lost or up again.
-- |
-- | Finally, its entire state can be provided by its parent.
-- | See the explanation for the `Output` data type.

data Query a
  = MessageReceived DNSManager.AnswerMessage a
  | ConnectionIsDown a
  | ConnectionIsUp a
  | ProvideState (Maybe State) a

type Slot = H.Slot Query Output

-- | `App.DomainListInterface` has no input.

type Input = Unit

-- | `App.DomainListInterface` has a single form to add a new domain.
-- | Only two possible inputs: the (sub)domain name and the selection of the TLD.

data NewDomainFormAction
  = INP_newdomain String
  | UpdateSelectedDomain String

-- | Possible component's actions are:
-- | - update the accepted domains (examples: netlib.re, codelib.re and example.com)
-- | - update the list of own domains
-- | - handle user inputs
-- | - add a new domain
-- | - remove a domain
-- | - TODO: show the zone content (in another component)

data Action
  = UpdateAcceptedDomains (Array String)
  | UpdateMyDomains (Array String)

  | HandleNewDomainInput NewDomainFormAction

  | NewDomainAttempt Event
  | RemoveDomain String
  | EnterDomain String

  | DeleteDomainModal String
  | CancelModal

  | Initialize
  | Finalize

-- | The form only has two elements:
-- | the subdomain name input and the selected TLD.

type NewDomainFormState
  = { new_domain      :: String
    , _errors         :: Array Validation.Error
    , selected_domain :: String
    }

-- | The entire component's state contains the form, accepted domains,
-- | the list of own domains and a boolean to know if the connection is up.

type State =
  { newDomainForm          :: NewDomainFormState
  , accepted_domains       :: Array String
  , my_domains             :: Array String

  , wsUp :: Boolean
  , active_modal :: Maybe String
  }

component :: forall m. MonadAff m => H.Component Query Input Output m
component =
  H.mkComponent
    { initialState
    , render
    , eval: H.mkEval $ H.defaultEval
        { initialize   = Just Initialize
        , handleAction = handleAction
        , handleQuery  = handleQuery
        , finalize     = Just Finalize
        }
    }

-- | Default available domain: netlib.re.

default_domain :: String
default_domain = "netlib.re"

initialState :: Input -> State
initialState _ =
  { newDomainForm: { new_domain: ""
                   , _errors: []
                   , selected_domain: default_domain
                   }
  , accepted_domains: [ default_domain ]
  , my_domains: [ ]
  , wsUp: true
  , active_modal: Nothing
  }

render :: forall m. State -> H.ComponentHTML Action () m
render { accepted_domains, my_domains, newDomainForm, wsUp, active_modal }
  = Bulma.section_small
      [ case wsUp of
          false -> Bulma.p "You are disconnected."
          true  -> case active_modal of
            Nothing -> Bulma.columns_
              [ Bulma.column_ [ Bulma.h3 "Add a domain!", render_add_domain_form]
              , Bulma.column_ [ Bulma.h3 "My domains"
                              , HH.ul_ $ map (\domain -> HH.li_ (domain_buttons domain)) my_domains
                              ]
              ]
            Just domain -> Bulma.modal "Deleting a domain"
              [warning_message domain] [modal_delete_button domain, modal_cancel_button]
      ]
  where
    modal_delete_button domain = Bulma.alert_btn "Delete the domain" (RemoveDomain domain)
    modal_cancel_button = Bulma.cancel_button CancelModal
    warning_message domain
      = HH.p [] [ HH.text $ "You are about to delete your domain '"
                            <> domain
                            <> "'. Are you sure you want to do this? This is "
                , HH.strong_ [ HH.text "irreversible" ]
                , HH.text "."
                ]

    domain_buttons domain
      = [ Bulma.alert_btn "delete" (DeleteDomainModal domain)
        , Bulma.btn domain (EnterDomain domain)
        ]

    render_add_domain_form = HH.form
      [ HE.onSubmit NewDomainAttempt ]
      [ Bulma.new_domain_field
          (HandleNewDomainInput <<< INP_newdomain)
          newDomainForm.new_domain
          [ HHE.onSelectedIndexChange domain_choice ]
          accepted_domains
      , HH.button
          [ HP.type_ HP.ButtonSubmit
          , HP.classes C.button
          ]
          [ HH.text "add a new domain!" ]
      , if A.length newDomainForm._errors > 0
        then HH.div_ $ map error_to_paragraph_label newDomainForm._errors
        else HH.div_ [ ]
      ]

    domain_choice :: Int -> Action
    domain_choice i
      = HandleNewDomainInput <<< UpdateSelectedDomain $ maybe default_domain (\x -> x) $ accepted_domains A.!! i

handleAction :: forall m. MonadAff m => Action -> H.HalogenM State Action () Output m Unit
handleAction = case _ of
  Initialize -> do
    H.raise $ AskState

  Finalize -> do
    state <- H.get
    H.raise $ StoreState state

  CancelModal -> do
    H.modify_ _ { active_modal = Nothing }

  UpdateAcceptedDomains domains -> do
    H.modify_ _ { accepted_domains = domains }

  UpdateMyDomains domains -> do
    H.modify_ _ { my_domains = domains }

  HandleNewDomainInput adduserinp -> do
    case adduserinp of
      INP_newdomain v -> do
        H.modify_ _ { newDomainForm { new_domain = v } }
        case v of
          "" -> H.modify_ _ { newDomainForm { _errors = [] } }
          _ -> case Validation.label v of
            Left arr -> H.modify_ _ { newDomainForm { _errors = arr } }
            Right _  -> H.modify_ _ { newDomainForm { _errors = [] } }
      UpdateSelectedDomain domain -> H.modify_ _ { newDomainForm { selected_domain = domain } }

  EnterDomain domain -> do
    H.raise $ ChangePageZoneInterface domain

  DeleteDomainModal domain -> do
    H.modify_ _ { active_modal = Just domain }

  RemoveDomain domain -> do
    message <- H.liftEffect $ DNSManager.serialize $ DNSManager.MkDeleteDomain { domain: domain }
    H.raise $ MessageToSend message
    H.raise $ Log $ SystemLog $ "Remove domain: " <> domain
    H.modify_ _ { active_modal = Nothing }

  NewDomainAttempt ev -> do
    H.liftEffect $ Event.preventDefault ev

    { newDomainForm } <- H.get
    let new_domain = build_new_domain newDomainForm.new_domain newDomainForm.selected_domain

    case newDomainForm._errors, new_domain of
      _, "" ->
        H.raise $ Log $ UnableToSend "You didn't enter the new domain!"
      [], _ -> do
        message <- H.liftEffect
          $ DNSManager.serialize
          $ DNSManager.MkNewDomain { domain: new_domain }
        H.raise $ MessageToSend message
        H.raise $ Log $ SystemLog $ "Add a new domain (" <> new_domain <> ")"
        handleAction $ HandleNewDomainInput $ INP_newdomain ""
      _, _ ->
        H.raise $ Log $ UnableToSend $ "You didn't enter a valid new domain"

handleQuery :: forall a m. MonadAff m => Query a -> H.HalogenM State Action () Output m (Maybe a)
handleQuery = case _ of

  ProvideState maybe_state a -> do
    case maybe_state of
      Nothing -> pure Nothing
      Just s  -> do
        H.put s
        pure (Just a)

  MessageReceived message a -> do
    case message of
      -- The authentication failed.
      (DNSManager.MkAcceptedDomains response) -> do
        handleAction $ UpdateAcceptedDomains response.domains
      (DNSManager.MkLogged response) -> do
        handleAction $ UpdateAcceptedDomains response.accepted_domains
        handleAction $ UpdateMyDomains response.my_domains
      (DNSManager.MkDomainAdded response) -> do
        { my_domains } <- H.get
        handleAction $ UpdateMyDomains (my_domains <> [response.domain])
      (DNSManager.MkDomainDeleted response) -> do
        { my_domains } <- H.get
        handleAction $ UpdateMyDomains $ A.filter ((/=) response.domain) my_domains
      _ -> H.raise $ Log $ ErrorLog $ "Message not handled in DomainListInterface."
    pure (Just a)

  ConnectionIsDown a -> do
    H.modify_ _ { wsUp = false }
    pure (Just a)

  ConnectionIsUp a -> do
    H.modify_ _ { wsUp = true }
    pure (Just a)

page_reload :: State -> DNSManager.AnswerMessage -> State
page_reload s1 message =
  case message of
    DNSManager.MkLogged response ->
      s1 { accepted_domains = response.accepted_domains
         , my_domains = response.my_domains
         }
    _ -> s1

build_new_domain :: String -> String -> String
build_new_domain sub tld
  | endsWith "." sub = sub <> tld
  | otherwise        = sub <> "." <> tld
