-- | `App.RegistrationInterface` is a registration interface.
-- | Registration requires a login, an email address and a password.
module App.Page.Registration where

import Prelude (Unit, bind, discard, pure, ($), (<<<), (<>), map, show)

import Data.Array as A
import Data.ArrayBuffer.Types (ArrayBuffer)
import Data.Maybe (Maybe(..), maybe)
import Data.Either (Either(..))
import Effect.Aff.Class (class MonadAff)
import Halogen as H
import Halogen.HTML as HH
import Halogen.HTML.Events as HE
import Halogen.HTML.Properties as HP
import Web.Event.Event as Event
import Web.Event.Event (Event)

import Bulma as Bulma

import App.Type.Email as Email
import App.Type.LogMessage
import App.Message.AuthenticationDaemon as AuthD

import App.Validation.Login as L
import App.Validation.Email as E
import App.Validation.Password as P

data Output
  = MessageToSend ArrayBuffer
  | Log LogMessage

-- | The component is informed when the connection went up or down.
data Query a
  = ConnectionIsDown a
  | ConnectionIsUp a

type Slot = H.Slot Query Output

type Input = Unit

data RegisterInput
  = REG_INP_login String
  | REG_INP_email String
  | REG_INP_pass String

data Action
  -- | Simply get the inputs from the form.
  = HandleRegisterInput RegisterInput
  -- | Validate inputs (login, email, password) then send the request
  -- | (via `SendRegistrationRequest`) or log errors.
  | ValidateInputs Event
  -- | Send the registration request to `dnsmanagerd`.
  -- | This action is automatically called from `ValidateInputs`.
  | SendRegistrationRequest

-- | The possible errors come from either the login, email or password input.
data Error
  = Login    (Array L.Error)
  | Email    (Array E.Error)
  | Password (Array P.Error)

-- | The whole registration form is composed of three strings: login, email and password.
type StateRegistrationForm = { login :: String, email :: String, pass :: String }

-- | State is composed of the registration form, the errors and an indication whether
-- | the websocket connection with `authd` is up or not.
type State =
  { registrationForm :: StateRegistrationForm
  , errors           :: Array Error
  , wsUp             :: Boolean
  }

initialState :: Input -> State
initialState _ =
  { registrationForm: { login: "", email: "", pass: "" }
  , errors: []
  , wsUp: true
  }

component :: forall m. MonadAff m => H.Component Query Input Output m
component =
  H.mkComponent
    { initialState
    , render
    , eval: H.mkEval $ H.defaultEval
        { handleAction = handleAction
        , handleQuery  = handleQuery
        }
    }

render :: forall m. State -> H.ComponentHTML Action () m
render { wsUp, registrationForm }
  = Bulma.section_small
      [ case wsUp of
          false -> Bulma.p "You are disconnected."
          true  -> Bulma.columns_ [ b registration_form ]
      ]

  where
    b e = Bulma.column_ [ Bulma.box e ]
    registration_form = [ Bulma.h3 "Register!", render_register_form ]

    should_be_disabled = (if wsUp then (HP.enabled true) else (HP.disabled true))

    render_register_form = HH.form
      [ HE.onSubmit ValidateInputs ]
      [ Bulma.box_input "loginREGISTER" "Login" "login"                -- title, placeholder
          (HandleRegisterInput <<< REG_INP_login)      -- action
          registrationForm.login                       -- value
          should_be_disabled                           -- condition
      , Bulma.box_input "emailREGISTER" "Email" "email@example.com"    -- title, placeholder
          (HandleRegisterInput <<< REG_INP_email)      -- action
          registrationForm.email                       -- value
          should_be_disabled                           -- condition
      , Bulma.box_password "passwordREGISTER" "Password" "password"       -- title, placeholder
          (HandleRegisterInput <<< REG_INP_pass)       -- action
          registrationForm.pass                        -- value
          should_be_disabled                           -- condition
      , HH.div_
          [ HH.button
              [ HP.style "padding: 0.5rem 1.25rem;"
              , HP.type_ HP.ButtonSubmit
              , (if wsUp then (HP.enabled true) else (HP.disabled true))
              ]
              [ HH.text "Send Message to Server" ]
          ]
      ]

handleAction :: forall m. MonadAff m => Action -> H.HalogenM State Action () Output m Unit
handleAction = case _ of
  HandleRegisterInput reginp -> do
    case reginp of
      REG_INP_login v -> H.modify_ _ { registrationForm { login = v } }
      REG_INP_email v -> H.modify_ _ { registrationForm { email = v } }
      REG_INP_pass  v -> H.modify_ _ { registrationForm { pass  = v } }

  -- Validate inputs (login, email, password) then send the request
  -- (via SendRegistrationRequest) or log errors.
  ValidateInputs ev -> do
    H.liftEffect $ Event.preventDefault ev

    { registrationForm } <- H.get
    let login = registrationForm.login
        email = registrationForm.email
        pass  = registrationForm.pass

    case login, email, pass of
      "", _, _ ->
        H.raise $ Log $ UnableToSend "Write your login!"

      _, "", _ ->
        H.raise $ Log $ UnableToSend "Write your email!"

      _, _, "" ->
        H.raise $ Log $ UnableToSend "Write your password!"

      _, _, _ -> do
        case L.login login, E.email email, P.password pass of
          Left errors, _, _         -> H.raise $ Log $ UnableToSend $ show_error $ Login errors
          _, Left errors, _         -> H.raise $ Log $ UnableToSend $ show_error $ Email errors
          _, _, Left errors         -> H.raise $ Log $ UnableToSend $ show_error $ Password errors
          Right _, Right _, Right _ -> handleAction $ SendRegistrationRequest

  SendRegistrationRequest -> do
    { registrationForm } <- H.get
    let { login, email, pass } = registrationForm
    message <- H.liftEffect $ AuthD.serialize $
      AuthD.MkRegister { login, email: Just (Email.Email email), password: pass }
    H.raise $ MessageToSend message
    H.raise $ Log $ SystemLog $ "Trying to register (login: " <> login <> ")"

show_error :: Error -> String
show_error = case _ of
  Login    arr -> "Error with the Login: "    <> (A.fold $ map show_error_login arr)
  Email    arr -> "Error with the Email: "    <> (A.fold $ map show_error_email arr)
  Password arr -> "Error with the Password: " <> (A.fold $ map show_error_password arr)

show_error_login :: L.Error -> String
show_error_login = case _ of
  L.ParsingError {error, position} ->
    "position " <> show position <> " " <> maybe "" string_error_login error

string_error_login :: L.LoginParsingError -> String
string_error_login = case _ of
  L.CannotParse         -> "cannot parse the login"
  L.CannotEntirelyParse -> "cannot entirely parse the login"
  L.Size min max n      -> "login size should be between "
                           <> show min <> " and " <> show max
                           <> " (currently: " <> show n <> ")"

show_error_email :: E.Error -> String
show_error_email = case _ of
  E.ParsingError {error, position} ->
    "position " <> show position <> " " <> maybe "" string_error_email error

string_error_email :: E.EmailParsingError -> String
string_error_email = case _ of
  E.CannotParse         -> "cannot parse the email"
  E.CannotEntirelyParse -> "cannot entirely parse the email"
  E.Size min max n      -> "email size should be between "
                           <> show min <> " and " <> show max
                           <> " (currently: " <> show n <> ")"

show_error_password :: P.Error -> String
show_error_password = case _ of
  P.ParsingError {error, position} ->
    "position " <> show position <> " " <> maybe "" string_error_password error

string_error_password :: P.PasswordParsingError -> String
string_error_password = case _ of
  P.CannotParse         -> "cannot parse the password"
  P.CannotEntirelyParse -> "cannot entirely parse the password"
  P.Size min max n      -> "password size should be between "
                           <> show min <> " and " <> show max
                           <> " (currently: " <> show n <> ")"

handleQuery :: forall a m. MonadAff m => Query a -> H.HalogenM State Action () Output m (Maybe a)
handleQuery = case _ of
  ConnectionIsDown a -> do
    H.modify_ _ { wsUp = false }
    pure (Just a)

  ConnectionIsUp a -> do
    H.modify_ _ { wsUp = true }
    pure (Just a)
